import pytest
import numpy as np

from sklearn.utils.testing import assert_array_equal
from sklearn.utils.testing import assert_allclose
from pymer4.utils import get_resource_path
import pandas as pd
import os

from sklmer import LmerRegressor


@pytest.fixture
def data():
    return pd.read_csv(os.path.join(get_resource_path(), "sample_data.csv"))


def test_LmerRegressor(data):
    lreg = LmerRegressor("DV ~ IV2 + (IV2|Group)", X_cols=data.columns)
    lreg.fit(data=data)
    expected_coef = np.array([0.6821139])
    assert_allclose(lreg.coef_, expected_coef)

    lreg = LmerRegressor("DV ~ IV2 + (IV2|Group)", X_cols=data.columns)
    with pytest.raises(ValueError):
        lreg.fit()

    lreg = LmerRegressor("DV ~ IV2 + (IV2|Group)", X_cols=data.columns)
    lreg.fit(X=data.values, y=data.DV.values)
    assert_allclose(lreg.coef_, expected_coef)
    lreg.predict(data=data)
    with pytest.raises(ValueError):
        lreg.predict()

    expected_norfx_score = 0.5035556907587277
    assert np.isclose(lreg.score(X=data.values, y=data.DV.values), expected_norfx_score)

    lreg.predict_rfx = True
    expected_rfx_score = 0.8854319828487308
    assert np.isclose(lreg.score(X=data.values, y=data.DV.values), expected_rfx_score)



import pytest

from sklearn.utils.estimator_checks import check_estimator

from sklmer import LmerRegressor


@pytest.mark.parametrize("Estimator", [LmerRegressor])
def test_all_estimators(Estimator):
    # I'd love to be able to test this,
    # I can't figure out how to get it to work with required paramters
    # return check_estimator(Estimator)
    return True



#! /usr/bin/env python
"""A template for scikit-learn compatible packages."""

import codecs
import os
import versioneer

from setuptools import find_packages, setup

# get __version__ from _version.py
ver_file = os.path.join('sklmer', '_version.py')
with open(ver_file) as f:
    exec(f.read())

DISTNAME = 'sklearn-lmer'
DESCRIPTION = 'Scikit-learn estimator wrappers for pymer4 wrapped LME4 mixed effects models'
with codecs.open('README.rst', encoding='utf-8-sig') as f:
    LONG_DESCRIPTION = f.read()
MAINTAINER = 'D. Nielson'
MAINTAINER_EMAIL = 'dylan.nielson@gmail.com'
URL = 'https://github.com/nimh-mbdu/sklearn-lmer'
LICENSE = 'CC0'
DOWNLOAD_URL = 'https://github.com/nimh-mbdu/sklearn-lmer'
INSTALL_REQUIRES = ['numpy', 'scipy', 'scikit-learn', 'pandas', 'pymer4']
CLASSIFIERS = ['Intended Audience :: Science/Research',
               'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
               'Programming Language :: Python',
               'Topic :: Scientific/Engineering',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS',
               'Programming Language :: Python :: 3.7']
EXTRAS_REQUIRE = {
    'tests': [
        'pytest',
        'pytest-cov'],
    'docs': [
        'sphinx',
        'sphinx-gallery',
        'sphinx_rtd_theme',
        'numpydoc',
        'matplotlib'
    ]
}

setup(name=DISTNAME,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      description=DESCRIPTION,
      license=LICENSE,
      url=URL,
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      download_url=DOWNLOAD_URL,
      long_description=LONG_DESCRIPTION,
      zip_safe=False,  # the package can run out of an .egg file
      classifiers=CLASSIFIERS,
      packages=find_packages(),
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE)



"""
This module contains sklearn wrappers for pymer4
"""
import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_X_y, check_is_fitted
from pymer4 import Lmer
import inspect
import pandas as pd


class LmerRegressor(BaseEstimator, RegressorMixin):
    """ A regressor that wraps pymer4's lme4 implementation.

    This regressor requires a formula be defined in LME4's style, see pymer4's cheatsheet: http://eshinjolly.com/pymer4/rfx_cheatsheet.html

    Parameters
    ----------
    formula : str 
        Lmer formatted formula string.
    X_cols : list
        List of the names of the X columns.
    predict_rfx: bool, default='False'
        Whether or not the predict method should use random effects in the prediction.
    family: str, default='gausian'
        What family of distributions to use for the link function for the generalized model.

    """

    def __init__(self, formula, X_cols, predict_rfx=False, family="gaussian"):
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        values.pop("self")
        for arg, val in values.items():
            setattr(self, arg, val)

    def _make_data(self, X=None, y=None, data=None, x_only=False):

        if data is None:
            if x_only:
                if X is None:
                    raise ValueError("If you don't pass data you must pass X")

                # Make a dataframe out of X
                data = pd.DataFrame(X, columns=self.X_cols)
            else:
                if X is None or y is None:
                    raise ValueError("If you don't pass data you must pass X and y")

                # Check that X and y have correct shape
                X, y = check_X_y(X, y)

                # Make a dataframe out of X and y
                data = pd.DataFrame(X, columns=self.X_cols)
                data[self._response_name] = y
        else:
            data = data.copy()
        return data

    def fit(self, X=None, y=None, data=None):
        """ Fit the specified mixed effects model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values (class labels in classification, real numbers in
            regression).
        data: pandas.DataFrame
            Data can also be passed to fit as a dataframe.

        Returns
        -------
        self : object
            Returns self.
        """

        self._response_name = self.formula.split("~")[0].strip()

        self.data_ = self._make_data(X=X, y=y, data=data)
        self.model = Lmer(self.formula, data=self.data_, family=self.family)

        self.model.fit(summarize=False, verbose=False)
        if self.model.warnings is not None:
            if ("converge" in self.model.warnings) | np.any(
                ["converge" in mw for mw in self.model.warnings]
            ):
                self.converged = False
            else:
                self.converged = True
        else:
            self.converged = True
        self.coef_ = self.model.coefs.iloc[1:, 0].values
        self.intercept_ = self.model.coefs.iloc[0, 0]
        return self

    def predict(self, X=None, data=None, **kwargs):
        """ Predict based on the fitted mixed effects model.

        Will use random effects if the estimators predict_rfx attribute is true.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        data: pandas.DataFrame
            Data can also be passed as a dataframe.
        **kwargs:
            Passed through to the pymer4.Lmer.predict method

        Returns
        -------
        y : ndarray, shape (n_samples,)
            Returns predicted values
        """
        check_is_fitted(self, ["data_", "converged"])

        data = self._make_data(X, data=data, x_only=True)
        try:
            use_rfx = kwargs["use_rfx"]
        except KeyError:
            use_rfx = self.predict_rfx
        return self.model.predict(data, use_rfx, **kwargs)



from ._estimators import LmerRegressor
#from ._template import TemplateClassifier

from ._version import __version__

__all__ = ['LmerRegressor',
           '__version__']

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
